%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% dictionary_learn_script.m 
% This scrips it an example script for basic sparse dictionary learning
% based on the Olshausen and Field, 1997 paper 'Sparse Coding with
% an Overcomplete Basis Set: a Strategy Employed by V1?'
% 
% Deviations from the model presented are: 
%   - use of L1 regularized BPDN solvers as well as options to use
%     hard-sparse MP-type solvers for the inference step. This allows
%     faster and more accurate solvutions for the sparse coefficients.
%   - Normalization of the dictionary elements to have unit norm at each
%     step rather than normalizing the variance of the coefficients. This
%     is the default. Also included is a method that normalizes the
%     dictionary elements by a Forbeneous norm.
%   
% Other optimizations:
% 
%   - Parallel for loops are used in the inference step to speed up
%     run-time. If the script is to be run on a cluster, ust the
%     createMatlabPoolJob() function to allow for this speedup. 
% 
% 
% Last Updated 6/3/2010 - Adam Charles
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Set Options

% Set multithreading to Ncores - 1 threads
% This speeds up matrix multiplication etc.
if matlabpool('size') ~= 0
    matlabpool CLOSE
end
matlabpool('open', 7)

% Random number seed
RandStream.setDefaultStream (RandStream('mt19937ar','seed',sum(100*clock)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Options

% Save Options
opts.save_name = '2010_07_13_44hyper_01.mat';  % Basis Save Name
opts.save_every = 500;              % How many iterations to save after

% Algorithm options
opts.data_type = 'vector';          % Type of data (vecror, square or cube)
opts.sparse_type = 'l1ls_nneg';     % Choose to use l1ls for sparsification OMP_qr
                                    % Options: l1ls, l1ls_nneg, sparsenet, MP
opts.grad_type = 'norm';            % Choose weather to include the Forb norm in E(a,D)
opts.nneg_dict = 1;                 % Make the basis values all positive

% Dictionary options
opts.n_elem = 44;                   % Number of dictionary elements

% Iteration numbers
opts.iters = 10000;                  % Number of learning iterations
opts.in_iter = 200;                 % Number of internal iterations
opts.GD_iters = 1;                  % Basis Gradient Descent iterations

% Specify Parameters
opts.step_size = 20;                 % Initial Step Size for Gradient Descent
opts.decay = 0.99998;                % Step size decay factor
opts.lambda = 0.01;                  % Lambda Value for Sparsity
opts.tol = 0.001;                   % Sparsification Tolerance

% Plotting Options
opts.bshow = 100;                     % Number of iterations to show basis
opts.disp_size = [11, 4];            % Basis display dimensions

%% Load Training Data

% Select data type. ''image'' or ''hyperspectral'' included here. Change to
% use other data types by simply loading the data into a matric called
% 'data_obj'. Check help file in learn_dictionary.m for data_type specific
% formats.

data_class = 'hyperspectral'; 

load('2010_007_13_HSIdata_20011018southend.mat')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Initialize Dictionary Elements
% To load an older dictionary, 

if strcmp(opts.data_type, 'vector')
    v_size = size(data_obj, 1);
elseif strcmp(opts.data_type, 'square')
    v_size = opts.bl_size.^2;
elseif strcmp(opts.data_type, 'cube')
    v_size = opts.dep_size*(opts.bl_size.^2);
else
    error('Unknown Data Type!! Choose ''vector'', ''square'' or ''cube''...')
end

dictionary_initial = initialize_dictionary(opts.n_elem, v_size, opts.nneg_dict);
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Run Learning Algorithm

% Actually run the learning algorithm
[dictionary_out] = learn_dictionary(training_data, dictionary_initial, @l1ls_nneg_wrapper, opts);
% [dictionary_out] = learn_dictionary_spmd(data_obj, dictionary_initial, opts);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Plotting

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Closing time

matlabpool CLOSE

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
